package net.sf.multiproject;
import bluej.extensions.*;
import bluej.extensions.event.PackageEvent;
import bluej.extensions.event.PackageListener;

import javax.swing.JFrame;
import javax.swing.JPanel;

import java.awt.Dimension;
import java.awt.Frame;
import java.awt.GridLayout;
import java.awt.Point;
import java.awt.Rectangle;

import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.ListSelectionModel;
import javax.swing.SwingUtilities;

import java.awt.GridBagLayout;
import javax.swing.JButton;
import java.awt.GridBagConstraints;
import java.awt.event.KeyEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Properties;
import java.util.Set;
import java.util.SortedSet;
import java.util.StringTokenizer;
import java.util.TreeSet;
import java.util.Vector;
//import java.util.concurrent.Semaphore;

import javax.swing.JTextField;
import javax.swing.JList;
import javax.swing.JScrollPane;
import javax.swing.JComboBox;
import javax.swing.event.PopupMenuEvent;
import javax.swing.event.PopupMenuListener;
import javax.swing.filechooser.FileFilter;
import javax.swing.JCheckBox;
import javax.swing.JTabbedPane;

/*
File: GUI.java
License: The MIT License

Copyright (c) 2008 Manuel Haim

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

/**
 * The graphical user interface.
 * Designed using the Visual Editor plugin for Eclipse.
 * @author Manuel Haim
 * @version 2008.04.18
 */
public class GUI {
	/** The BlueJ instance (null if GUI is running stand-alone without BlueJ). */
	private BlueJ bluej = null;  //  @jve:decl-index=0:
	
	/** Properties for current workspace. */
	private Properties workspaceProps = new Properties();  //  @jve:decl-index=0:
	
	/** Global properties for the Multi-Project Workspace (only set if {@link #bluej}==null). */
	private Properties globalProps = null;
	
	/** Path to last opened workspace. */
	private String lastRefreshedWorkspace = null;
	private boolean terminated = false;
	
	/** A directory chooser. */
	private JFileChooser fileChooser;
	private JFileChooser fileChooser2;
	
	/** The package listener. */
	private PackageListener pkgListener = null;
	
	/** A semaphore ensuring ComboBox changes are ignored while last change is being processed
	 * (which will reorganize the ComboBox items and thus fire change events). */
	private final Semaphore comboChange = new Semaphore(1);  //  @jve:decl-index=0:
	
	/** A semaphore ensuring that opening/closing a project by Multi-Project Workspace will be ignored
	 * by BlueJ opening/closing event listener. */
	private final Semaphore projectChange = new Semaphore(1);  //  @jve:decl-index=0:
	
	/** Another semaphore ensuring that Multi-Project Workspace will only open/close one project at a time
	 * (see {@link #projectChange}).
	 */
	private final Semaphore projectChangeMutex = new Semaphore(1);  //  @jve:decl-index=0:
	
	/** Length of workspace history (number of ComboBox entries). */
	private int maxWorkspaces = 5;
	
	
	private JFrame jFrame = null;  //  @jve:decl-index=0:visual-constraint="181,-1"
	private JPanel jContentPane = null;
	private JPanel jPanel = null;
	private JTextField jTextField = null;
	private JPanel jPanel2 = null;
	private JPanel jPanel3 = null;
	private JButton jButton = null;
	private JList jList = null;
	private JScrollPane jScrollPane = null;
	private JButton jButton1 = null;
	private JButton jButton2 = null;
	private JButton jButton3 = null;
	private JPanel jPanel5 = null;
	private JButton jButton4 = null;
	private JButton jButton5 = null;
	private JComboBox jComboBox = null;
	private JPanel jPanel6 = null;
	private JButton jButton6 = null;
	private JButton jButton7 = null;
	private JPanel jPanel4 = null;
	private JButton jButton8 = null;
	private JPanel jPanel7 = null;
	private JButton jButton9 = null;
	private JTabbedPane jTabbedPane = null;
	private JPanel jPanel1 = null;
	private JPanel jPanel8 = null;
	private JPanel jPanel9 = null;
	private JPanel jPanel10 = null;
	private JCheckBox jCheckBox = null;
	private JPanel jPanel11 = null;
	private JPanel jPanel12 = null;
	private JPanel jPanel13 = null;
	private JCheckBox jCheckBox1 = null;
	private JButton jButton10 = null;
	private JButton jButton11 = null;
	private JButton jButton12 = null;
	private JButton jButton13 = null;
	private JTabbedPane jTabbedPane1 = null;
	private JPanel jPanel14 = null;
	private JPanel jPanel15 = null;
	private JPanel jPanel16 = null;
	private JPanel jPanel17 = null;
	private JPanel jPanel18 = null;
	private JButton jButton14 = null;
	private JButton jButton15 = null;
	private JCheckBox jCheckBox2 = null;
	private JButton jButton16 = null;
	private JButton jButton17 = null;
	private JPanel jPanel19 = null;
	private JCheckBox jCheckBox3 = null;
	private JPanel jPanel20 = null;
	private JCheckBox jCheckBox4 = null;

	private JCheckBox jCheckBox5 = null;  //  @jve:decl-index=0:visual-constraint="79,330"

	/**
	 * This method initializes jFrame	
	 * 	
	 * @return javax.swing.JFrame	
	 */
	private JFrame getJFrame() {
		if (jFrame == null) {
			jFrame = new JFrame();
			jFrame.setSize(new Dimension(180, 480));
			jFrame.setTitle("Workspace");
			jFrame.setContentPane(getJContentPane());
			jFrame.addWindowListener(new java.awt.event.WindowAdapter() {
				public void windowClosing(java.awt.event.WindowEvent e) {
					//System.out.println("windowClosing()");
					//JOptionPane.showMessageDialog(null, "Window closing.");
				}
			});
			//jList.setModel(new MyListModel(new String[] {"a", "b", "c"}));

		}
		//jFrame.pack(); //set size to best fit
		return jFrame;
	}

	/**
	 * This method initializes jContentPane	
	 * 	
	 * @return javax.swing.JPanel	
	 */
	private JPanel getJContentPane() {
		if (jContentPane == null) {
			jContentPane = new JPanel();
			jContentPane.setLayout(new BoxLayout(jContentPane, BoxLayout.PAGE_AXIS));
			//jContentPane.setBorder(BorderFactory.createEmptyBorder(5,5,5,5));
			jContentPane.add(getJPanel5());
			jContentPane.add(getJPanel6());
			jContentPane.add(getJPanel1());
		}
		return jContentPane;
	}

	/**
	 * This method initializes jPanel	
	 * 	
	 * @return javax.swing.JPanel	
	 */
	private JPanel getJPanel() {
		if (jPanel == null) {
			jPanel = new JPanel();
			jPanel.add(getJButton7());
		}
		return jPanel;
	}


	/**
	 * This method initializes jTextField	
	 * 	
	 * @return javax.swing.JTextField	
	 */
	private JTextField getJTextField() {
		if (jTextField == null) {
			jTextField = new JTextField();
			jTextField.setColumns(10);
			jTextField.setToolTipText("Enter path to import dir for Unpack (workspace-specific, must contain Java Archive files).");
			jTextField.addActionListener(new java.awt.event.ActionListener() {
				public void actionPerformed(java.awt.event.ActionEvent e) {
					myRefreshList();
				}
			});
		}
		return jTextField;
	}

	/**
	 * This method initializes jPanel2	
	 * 	
	 * @return javax.swing.JPanel	
	 */
	private JPanel getJPanel2() {
		if (jPanel2 == null) {
			jPanel2 = new JPanel();
			jPanel2.setLayout(new GridLayout(0,2));
			jPanel2.add(getJButton15());
			jPanel2.add(getJButton10());
			jPanel2.add(getJButton13());
			jPanel2.add(getJButton12());
			jPanel2.add(getJButton3());
			jPanel2.add(getJButton11());
			jPanel2.add(getJButton8());
			jPanel2.add(getJButton1());
		}
		return jPanel2;
	}

	/**
	 * This method initializes jPanel3	
	 * 	
	 * @return javax.swing.JPanel	
	 */
	private JPanel getJPanel3() {
		if (jPanel3 == null) {
			jPanel3 = new JPanel();
			jPanel3.setLayout(new BoxLayout(jPanel3, BoxLayout.PAGE_AXIS));
			jPanel3.add(getJPanel8());
			jPanel3.add(getJTabbedPane1());
		}
		return jPanel3;
	}

	/**
	 * This method initializes jButton	
	 * 	
	 * @return javax.swing.JButton	
	 */
	private JButton getJButton() {
		if (jButton == null) {
			jButton = new JButton();
			jButton.setText("Unpack");
			jButton.setToolTipText("Import/Unpack selected project from import dir (no open, ask before overwrite).");
			jButton.addActionListener(new java.awt.event.ActionListener() {
				public void actionPerformed(java.awt.event.ActionEvent e) {
					if(jList.getSelectedIndex()==-1) return;
					File file = ((MyFile) jList.getSelectedValue()).getFile();
					boolean deleted = false;
					if(file.isDirectory()) {
						if(!myUnpackJarfileExists(file) && myGetUnpackSubdir(file)==null) {
							JOptionPane.showMessageDialog(null, "Cannot re-import project:\nThere is no import file or subdir for this project.");
							return;
						}
						if(JOptionPane.showConfirmDialog(null, "Really overwrite project by re-import?\n" + file.getAbsolutePath(), "Import project", JOptionPane.OK_CANCEL_OPTION)
						== JOptionPane.OK_OPTION) {
							try {
								projectChangeMutex.acquire();
								projectChange.acquire();
								if(bluej!=null) deleted = myFindDelete(file);
							} catch (InterruptedException e1) {
							} finally {
								projectChange.release();
								projectChangeMutex.release();
							}
							if(!deleted) deleted = FileHelpers.deleteDir(file);
						}
					} else deleted = true;
					if(deleted) {
						myUnpack();
						myRefreshList();
					}
				}

			});
		}
		return jButton;
	}

	/**
	 * This method initializes jList	
	 * 	
	 * @return javax.swing.JList	
	 */
	private JList getJList() {
		if (jList == null) {
			jList = new JList();
			jList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
			jList.setToolTipText("Double-click or press Enter to open a project (*unpack if non-existing).");
			jList.addMouseListener(new java.awt.event.MouseAdapter() {
				public void mouseClicked(java.awt.event.MouseEvent e) {
					if(e.getClickCount()==2) try {
						projectChangeMutex.acquire();
						projectChange.acquire();
						mySwitch();
					} catch (InterruptedException e1) {
					} finally {
						projectChange.release();
						projectChangeMutex.release();
					}
				}
			});
			jList.addKeyListener(new java.awt.event.KeyListener() {
				public void keyTyped(java.awt.event.KeyEvent e) {
				}
				public void keyPressed(java.awt.event.KeyEvent e) {
				}
				public void keyReleased(java.awt.event.KeyEvent e) {
					if(e.getKeyCode() == KeyEvent.VK_ENTER) try {
						projectChangeMutex.acquire();
						projectChange.acquire();
						mySwitch();
					} catch (InterruptedException e1) {
					} finally {
						projectChange.release();
						projectChangeMutex.release();
					}
				}
			});
		}
		return jList;
	}

	/**
	 * This method initializes jScrollPane	
	 * 	
	 * @return javax.swing.JScrollPane	
	 */
	private JScrollPane getJScrollPane() {
		if (jScrollPane == null) {
			jScrollPane = new JScrollPane();
			jScrollPane.setViewportView(getJList());
		}
		return jScrollPane;
	}

	/**
	 * This method initializes jButton1	
	 * 	
	 * @return javax.swing.JButton	
	 */
	private JButton getJButton1() {
		if (jButton1 == null) {
			jButton1 = new JButton();
			jButton1.setText("Refresh");
			jButton1.setToolTipText("Re-Scan workspace (*and import dir) for project subdirs.");
			jButton1.addActionListener(new java.awt.event.ActionListener() {
				public void actionPerformed(java.awt.event.ActionEvent e) {
					myRefreshList();
				}
			});
		}
		return jButton1;
	}

	/**
	 * This method initializes jButton2	
	 * 	
	 * @return javax.swing.JButton	
	 */
	private JButton getJButton2() {
		if (jButton2 == null) {
			jButton2 = new JButton();
			jButton2.setText("Open++");
			jButton2.setToolTipText("Open selected project without closing the others (*unpack if non-existing).");
			jButton2.addActionListener(new java.awt.event.ActionListener() {
				public void actionPerformed(java.awt.event.ActionEvent e) {
					if(jList.getSelectedIndex()==-1) return;
					try {
						projectChangeMutex.acquire();
						projectChange.acquire();
						myUnpack();
						myOpen();
					} catch (InterruptedException e1) {
					} finally {
						projectChange.release();
						projectChangeMutex.release();
					}
				}

			});
		}
		return jButton2;
	}

	/**
	 * This method initializes jButton3	
	 * 	
	 * @return javax.swing.JButton	
	 */
	private JButton getJButton3() {
		if (jButton3 == null) {
			jButton3 = new JButton();
			jButton3.setText("Delete");
			jButton3.setToolTipText("Delete selected project from workspace (ask before).");
			jButton3.addActionListener(new java.awt.event.ActionListener() {
				public void actionPerformed(java.awt.event.ActionEvent e) {
					if(jList.getSelectedIndex()==-1) return;
					File file = ((MyFile) jList.getSelectedValue()).getFile();
					boolean deleted = false;
					if(file.isDirectory()) {
						if(JOptionPane.showConfirmDialog(null, "Really delete?\n" + file.getAbsolutePath(), "Delete project", JOptionPane.YES_NO_OPTION)
						== JOptionPane.YES_OPTION) {
							try {
								projectChangeMutex.acquire();
								projectChange.acquire();
								if(bluej!=null) deleted = myFindDelete(file);
							} catch (InterruptedException e1) {
							} finally {
								projectChange.release();
								projectChangeMutex.release();
							}
							if(!deleted) deleted = FileHelpers.deleteDir(file);
						} else return;
					}
					if(deleted) {
						myRefreshList();
					} else {
						JOptionPane.showMessageDialog(null, "Could not delete project:\n" + file.getAbsolutePath());
					}
				}
			});

		}
		return jButton3;
	}

	/**
	 * This method initializes jPanel5	
	 * 	
	 * @return javax.swing.JPanel	
	 */
	private JPanel getJPanel5() {
		if (jPanel5 == null) {
			jPanel5 = new JPanel();
			//jPanel5.setLayout(new GridLayout());
			jPanel5.add(getJButton6());
		}
		return jPanel5;
	}

	/**
	 * This method initializes jButton4	
	 * 	
	 * @return javax.swing.JButton	
	 */
	private JButton getJButton4() {
		if (jButton4 == null) {
			jButton4 = new JButton();
			jButton4.setText("Pack");
			jButton4.setToolTipText("Pack selected project to Java Archive (ask before overwrite).");
			jButton4.addActionListener(new java.awt.event.ActionListener() {
				public void actionPerformed(java.awt.event.ActionEvent e) {
					if(jList.getSelectedIndex()==-1) return;
					myPack();
				}
			});
		}
		return jButton4;
	}

	/**
	 * This method initializes jButton5	
	 * 	
	 * @return javax.swing.JButton	
	 */
	private JButton getJButton5() {
		if (jButton5 == null) {
			jButton5 = new JButton();
			jButton5.setText("Pack all");
			jButton5.setToolTipText("Pack all projects to Java Archives (ask before overwrite).");
			jButton5.addActionListener(new java.awt.event.ActionListener() {
				public void actionPerformed(java.awt.event.ActionEvent e) {
					myPackAll();
				}
			});
		}
		return jButton5;
	}

	/**
	 * This method initializes jComboBox	
	 * 	
	 * @return javax.swing.JComboBox	
	 */
	private JComboBox getJComboBox() {
		if (jComboBox == null) {
			jComboBox = new JComboBox();
			jComboBox.setEditable(true);
			jComboBox.setToolTipText("Enter path to workspace (dir with projects).");
			jComboBox.addItem(".");
			jComboBox.addItemListener(new java.awt.event.ItemListener() {
				public void itemStateChanged(java.awt.event.ItemEvent e) {
					if(e.getID()==java.awt.event.ItemEvent.ITEM_STATE_CHANGED
					&& e.getStateChange()==java.awt.event.ItemEvent.SELECTED
					&& !jComboBox.isPopupVisible()) {
						comboBoxChange(e.getItem());
					}
				}
			});

			jComboBox.addPopupMenuListener(new PopupMenuListener() {

				public void popupMenuCanceled(PopupMenuEvent e) {
				}

				public void popupMenuWillBecomeInvisible(PopupMenuEvent e) {
					comboBoxChange(jComboBox.getSelectedItem());
				}

				public void popupMenuWillBecomeVisible(PopupMenuEvent e) {
				}
				
			});
		}
		return jComboBox;
	}

	/**
	 * This method initializes jPanel6	
	 * 	
	 * @return javax.swing.JPanel	
	 */
	private JPanel getJPanel6() {
		if (jPanel6 == null) {
			jPanel6 = new JPanel();
			jPanel6.setLayout(new GridLayout());
			jPanel6.add(getJComboBox());
		}
		return jPanel6;
	}

	/**
	 * Open plugin window after menu click (within BlueJ).
	 */
	protected void startup(BlueJ bluej) {
		try {
			MultiProject.startedSem.acquire();
			if(MultiProject.started) {
				jFrame.requestFocusInWindow();
				jFrame.toFront();
			}
			if(!MultiProject.started) {
				this.bluej = bluej;
				javax.swing.SwingUtilities.invokeLater(new Runnable() {
					/** Recent size and position of the BlueJ window (will be used to position the BlueJ window
					 * for a newly-opened project or package) . */
					private java.awt.Rectangle lastBounds = null;
					
					/** Name of recent package. */
					private String lastBoundsPkgName = null;
					private Semaphore lastBoundsMutex = new Semaphore(1);
					
		            public void run() {
						getJFrame();
						//jFrame.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
						jFrame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
						jFrame.addWindowListener(
							new WindowAdapter() {
								public void windowClosing(WindowEvent e) {
									terminate();
								}
							}
						);
						loadConfig();
						loadWorkspaceConfig();
						myRefreshList();

						
						// Try to find the currently open project in list:
						myFindOpen();
						
						GUI.this.bluej.addPackageListener(pkgListener = new PackageListener() {
							/** Number of running packageClosing and packageOpening methods. */
							private int processes = 0;
							
							/** A semaphore for the tryAcquire and release methods. */
							private Semaphore mutex = new Semaphore(1);
							
							/**
							 * Acquire the projectChange semaphore to ignore changes caused by
							 * Multi-Project Workspace (reader/writer problem with reader priority,
							 * this is the reader, but cancel reading attempt if a writer is active).
							 * @return true if acquired, false otherwise.
							 */
							private boolean tryAcquire() {
								boolean acquired = false;
								try {
									mutex.acquire();
									processes++;
									if(processes==1) acquired = GUI.tryAcquire(projectChange);
								} catch (InterruptedException e) {
								} finally {
									if(!acquired) processes--;
									mutex.release();
								}
								return acquired;
							}

							/**
							 * Release the projectChange semaphore.
							 */
							private void release() {
								try {
									mutex.acquire();
									processes--;
									if(processes==0) projectChange.release();
								} catch (InterruptedException e) {
								} finally {
									mutex.release();
								}
							}

							public void packageClosing(PackageEvent arg0) {
								Frame frame = null;
								try {
									lastBoundsMutex.acquire();
									// In BlueJ 2.2.1, the closing package frame will not be
									// disposed if its getFrame() method is called here...
									frame = arg0.getPackage().getFrame();
									lastBounds = frame.getBounds();
									lastBoundsPkgName = arg0.getPackage().getName();
								} catch (InterruptedException e) {
								} catch (Exception e) {
								} finally {
									lastBoundsMutex.release();
									
									if(frame!=null) {
										// ... so we have to dispose the frame manually!
										final Frame f = frame;
										//EventQueue.invokeLater(new Runnable() {
										SwingUtilities.invokeLater(new Runnable() {
											public void run() {
												// Remember to keep the last frame open. It will be identified by
												// having no ":" in its title (it may be just "BlueJ"
												// instead of "BlueJ: <Project> [<Package>]").
												if(f.isVisible()) {
													if(f.getTitle().indexOf(":")!=-1) {
														f.dispose();
													} else {
														// the following may interfere with packageOpening()
														// (resulting in two threads trying to set the window position)
														
														/*
														// set size and position
														if(jCheckBox5.isSelected() || jCheckBox3.isSelected() || jCheckBox4.isSelected()) {
															Thread t = new MyPkgPositioner(f, "", lastBounds, lastBoundsPkgName);
															t.start();
														}
														*/
													}
												}
											}
										});
									}
								}
								
								


								// try to acquire the semaphore
								if(!tryAcquire()) return;
								
								// do something here

								// release the semaphore
								release();
							}


							
							
							public void packageOpened(PackageEvent arg0) {
								try {
									lastBoundsMutex.acquire();
									// look for previously active package window:
									BPackage currentPkg = GUI.this.bluej.getCurrentPackage();
									if(currentPkg!=null) try {
										// ensure that it is different from the opening package
										if(!currentPkg.getDir().equals(arg0.getPackage().getDir())
										|| !currentPkg.getName().equals(arg0.getPackage().getName())) {
											lastBounds = currentPkg.getFrame().getBounds();
											lastBoundsPkgName = currentPkg.getName();
										}
									} catch (ProjectNotOpenException e) {
									} catch (PackageNotFoundException e) {
									}
									
									// set size and position
									if(jCheckBox5.isSelected() || jCheckBox3.isSelected() || jCheckBox4.isSelected()) {
										Thread t = new MyPkgPositioner(arg0.getPackage(), lastBounds, lastBoundsPkgName);
										t.start();
									}
								} catch (InterruptedException e) {
								} finally {
									lastBoundsMutex.release();
								}

								// try to acquire the semaphore
								if(!tryAcquire()) return;

								// find in project list
								myFindOpen();

								// release the semaphore
								release();
							}
						});
						
						
						jFrame.setVisible(true);
						
						// The following code will position the BlueJ window next to
						// the extension window, but only if no project is open and
						// if one of the positioning options is checked.
						// This may be irritating to users.
						/*
						if(GUI.this.bluej.getOpenProjects().length==0) try {
							lastBoundsMutex.acquire();
							lastBounds = GUI.this.bluej.getCurrentFrame().getBounds();
							lastBounds.x = jFrame.getBounds().x + jFrame.getBounds().width + 2;
							lastBounds.y = jFrame.getBounds().y;
							lastBoundsPkgName = "";
							if(jCheckBox5.isSelected() || jCheckBox3.isSelected() || jCheckBox4.isSelected()) {
								Frame f = GUI.this.bluej.getCurrentFrame();
								if(jCheckBox5.isSelected()) f.setSize(lastBounds.width, lastBounds.height);
								if(jCheckBox3.isSelected() || jCheckBox4.isSelected()) f.setLocation(lastBounds.x, lastBounds.y);
								f.validate();
							}
						} catch (InterruptedException e) {
						} finally {
							lastBoundsMutex.release();
						}
						*/
						
						jFrame.requestFocusInWindow();
						jFrame.toFront();
		            }
		        });
				MultiProject.started = true;
				terminated = false;
			}
			MultiProject.startedSem.release();
		} catch(InterruptedException e) {}
	}
	
	/**
	 * Startup a closable instance (as stand-alone application without BlueJ).
	 */
	protected void startupClosable() {
		javax.swing.SwingUtilities.invokeLater(new Runnable() {
            public void run() {
				getJFrame();
				jFrame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
				jFrame.addWindowListener(new WindowListener() {

					public void windowActivated(WindowEvent e) {
					}

					public void windowClosed(WindowEvent e) {
					}

					public void windowClosing(WindowEvent e) {
						saveWorkspaceConfig();
						saveConfig();
					}

					public void windowDeactivated(WindowEvent e) {
					}

					public void windowDeiconified(WindowEvent e) {
					}

					public void windowIconified(WindowEvent e) {
					}

					public void windowOpened(WindowEvent e) {
					}
					
				});

				loadConfig();
				loadWorkspaceConfig();
				myRefreshList();
				jFrame.setVisible(true);
				
            }
        });
	}
	
	/**
	 * This method is called on window termination (within BlueJ).
	 */
	protected void terminate() {
		// Try to free resources.
		try {
			MultiProject.startedSem.acquire();
			if(MultiProject.started) {
				if(pkgListener!=null && bluej!=null) {
					bluej.removePackageListener(pkgListener);
					pkgListener = null;
				}
				saveWorkspaceConfig();
				saveConfig();		
				MultiProject.started = false;
				terminated = true;
			}
			jFrame.dispose();
			jFrame = null;
			MultiProject.startedSem.release();
		} catch(InterruptedException e) {}
	}
	
	protected boolean isTerminated() {
		return terminated;
	}
	
	/**
	 * Load config values from BlueJ properties file or local Config file.
	 */
	private void loadConfig() {
		/*
		File dir = MultiProject.this.bluej.getUserConfigDir();
		*/
		
		if(bluej==null) {
			globalProps = new Properties();
			File dir = (new File(".")).getAbsoluteFile();
			if(dir.isDirectory()) {
				String filename = FileHelpers.addFileSeparator(dir.getAbsolutePath()) + MultiProject.GLOBAL_SETTINGS_FILE;
				File configFile = new File(filename);
				if(configFile.exists()) {
					try {
						InputStream is = new FileInputStream(configFile);
						globalProps.load(is);
						is.close();
					} catch(Exception e) {}
				}
			}
		}
		
		int maxpaths = maxWorkspaces;
		try {
			maxpaths = Integer.parseInt(getGlobalProperty(MultiProject.PROPERTY_HISTORY_SIZE, ""+maxWorkspaces));
		} catch (NumberFormatException e) {
			maxpaths = maxWorkspaces;
		}
		maxWorkspaces = Math.max(1, maxpaths);

		jComboBox.removeAllItems();
		StringTokenizer st = new StringTokenizer(
			getGlobalProperty(MultiProject.PROPERTY_WORKSPACE, ""),
			File.pathSeparator
		);
		while(st.hasMoreTokens() && jComboBox.getItemCount() < maxWorkspaces) {
			String token = st.nextToken();
			if(!token.equals("")) jComboBox.addItem(token);
		}
		if(jComboBox.getItemCount()>0) jComboBox.setSelectedIndex(0);
	}
	
	/**
	 * Get a global property from BlueJ or Config file.
	 * @param key Key.
	 * @param defaultValue Default value.
	 * @return Value.
	 */
	private String getGlobalProperty(String key, String defaultValue) {
		if(bluej!=null) return bluej.getExtensionPropertyString(key, defaultValue);
		if(globalProps!=null) return globalProps.getProperty(key, defaultValue);
		return defaultValue;
	}
	
	/**
	 * Store a global property to BlueJ or Config file.
	 * @param key Key.
	 * @param value Value.
	 */
	private void setGlobalProperty(String key, String value) {
		if(bluej!=null) bluej.setExtensionPropertyString(key, value);
		if(globalProps!=null) globalProps.setProperty(key, value);
	}

	/**
	 * This is called whenever the ComboBox selection changes.
	 * @param newItem The newly selected item.
	 */
	private void comboBoxChange(Object newItem) {
		if(newItem==null) newItem = "";
		
		// use a semaphore to ignore changes done by this method:
		if(tryAcquire(comboChange)) {
			// Save previous workspace settings
			saveWorkspaceConfig();

			// Re-order workspace history
			Vector v = new Vector();
			v.add(newItem);
			for(int i=0; i<jComboBox.getItemCount() && v.size()<maxWorkspaces; i++) {
				if(!v.contains(jComboBox.getItemAt(i))) v.add(jComboBox.getItemAt(i));
			}
			jComboBox.removeAllItems();
			for(java.util.Iterator it = v.iterator(); it.hasNext(); ) {
				jComboBox.addItem(it.next());
			}
			jComboBox.setSelectedItem(newItem);
			
			// Load workspace settings
			boolean validWS = loadWorkspaceConfig();
			
			// Refresh workspace project list
			if(validWS) myRefreshList();
			
			comboChange.release();
		}
	}

	/**
	 * Load workspace specific configuration.
	 * @return true if valid workspace, false otherwise.
	 */
	private boolean loadWorkspaceConfig() {
		Object item = jComboBox.getSelectedItem();
		if(item==null || item.toString().equals("")) return false;

		String workspacePath = item.toString();
		
		lastRefreshedWorkspace = workspacePath;
		workspaceProps = new Properties();
		
		File dir = (new File(workspacePath)).getAbsoluteFile();
		if(dir.isDirectory()) {
			String filename = FileHelpers.addFileSeparator(dir.getAbsolutePath()) + MultiProject.WORKSPACE_SETTINGS_FILE;
			File configFile = new File(filename);
			if(configFile.exists()) {
				try {
					InputStream is = new FileInputStream(configFile);
					workspaceProps.load(is);
					is.close();
				} catch(Exception e) {}
			}
		}
		
		jTextField.setText(workspaceProps.getProperty(MultiProject.PROPERTY_IN, ""));
		jCheckBox.setSelected(workspaceProps.getProperty(MultiProject.PROPERTY_FILESFROMSUBDIRS, "false").equals("true"));
		jCheckBox2.setSelected(workspaceProps.getProperty(MultiProject.PROPERTY_JARSFROMSUBDIRS, "false").equals("true"));
		jCheckBox1.setSelected(workspaceProps.getProperty(MultiProject.PROPERTY_NOJAVADOC, "false").equals("true"));
		jCheckBox5.setSelected(workspaceProps.getProperty(MultiProject.PROPERTY_FIXEDWINDOWSIZE, "false").equals("true"));
		jCheckBox3.setSelected(workspaceProps.getProperty(MultiProject.PROPERTY_FIXEDWINDOWPOSITION, "false").equals("true"));
		jCheckBox4.setSelected(workspaceProps.getProperty(MultiProject.PROPERTY_CASCADINGWINDOWPOSITION, "false").equals("true"));
		
		return true;
	}
	
	/**
	 * Save workspace specific configuration.
	 */
	private void saveWorkspaceConfig() {
		if(lastRefreshedWorkspace==null) return;
		String workspacePath = lastRefreshedWorkspace;
		
		workspaceProps.setProperty(MultiProject.PROPERTY_IN, jTextField.getText());
		workspaceProps.setProperty(MultiProject.PROPERTY_FILESFROMSUBDIRS, Boolean.toString(jCheckBox.isSelected()));
		workspaceProps.setProperty(MultiProject.PROPERTY_JARSFROMSUBDIRS, Boolean.toString(jCheckBox2.isSelected()));
		workspaceProps.setProperty(MultiProject.PROPERTY_NOJAVADOC, Boolean.toString(jCheckBox1.isSelected()));
		workspaceProps.setProperty(MultiProject.PROPERTY_FIXEDWINDOWSIZE, Boolean.toString(jCheckBox5.isSelected()));
		workspaceProps.setProperty(MultiProject.PROPERTY_FIXEDWINDOWPOSITION, Boolean.toString(jCheckBox3.isSelected()));
		workspaceProps.setProperty(MultiProject.PROPERTY_CASCADINGWINDOWPOSITION, Boolean.toString(jCheckBox4.isSelected()));
		
		File dir = (new File(workspacePath)).getAbsoluteFile();
		if(dir.isDirectory()) {
			String filename = FileHelpers.addFileSeparator(dir.getAbsolutePath()) + MultiProject.WORKSPACE_SETTINGS_FILE;
			File configFile = new File(filename);
			if(!configFile.exists() || configFile.delete()) {
				try {
					OutputStream os = new FileOutputStream(configFile);
					workspaceProps.store(os, "Multi-Project Workspace " + MultiProject.MULTIPROJECT_VERSION);
					os.close();
				} catch(Exception e) {}
			}
		}
	}

	/**
	 * Save config values to BlueJ properties file.
	 */
	private void saveConfig() {
		/*
		File dir = bluej.getUserConfigDir();
		*/

		String workspaces = "";
		for(int i=0; i<jComboBox.getItemCount(); i++) {
			workspaces += jComboBox.getItemAt(i);
			if(i+1<jComboBox.getItemCount()) workspaces += File.pathSeparator;
		}
		setGlobalProperty(MultiProject.PROPERTY_WORKSPACE, workspaces);
		setGlobalProperty(MultiProject.PROPERTY_HISTORY_SIZE, ""+maxWorkspaces);
		if(bluej!=null) bluej.setExtensionPropertyString(MultiProject.PROPERTY_STARTUP, "false");
		
		if(bluej==null) {
			File dir = (new File(".")).getAbsoluteFile();
			if(dir.isDirectory()) {
				String filename = FileHelpers.addFileSeparator(dir.getAbsolutePath()) + MultiProject.GLOBAL_SETTINGS_FILE;
				File configFile = new File(filename);
				if(!configFile.exists() || configFile.delete()) {
					try {
						OutputStream os = new FileOutputStream(configFile);
						globalProps.store(os, "Multi-Project Workspace " + MultiProject.MULTIPROJECT_VERSION + " stand-alone settings");
						os.close();
					} catch(Exception e) {}
				}
			}
			
		}
	}
	
	/**
	 * A class trying to set the position of a package window.
	 * @author Manuel Haim
	 */
	class MyPkgPositioner extends Thread {
		Frame pkgFrame;
		String pkgName;
		int x;
		int y;
		int width;
		int height;
		String boundsPkgName;
		int attempts;
		
		/**
		 * Constructor (may be called by every opening window).
		 * @param pkg The package to position.
		 * @param bounds The bounds of the previous package.
		 * @param boundsPkgName The name of the previous package.
		 */
		public MyPkgPositioner(BPackage pkg, Rectangle bounds, String boundsPkgName) {
			this(null, null, bounds, boundsPkgName);
			attempts = 10;
			try {
				this.pkgFrame = pkg.getFrame();
				this.pkgName = pkg.getName();
			} catch (ProjectNotOpenException e) {
			} catch (PackageNotFoundException e) {
			}
		}
		
		/**
		 * Constructor (may be called by last closing window).
		 * @param pkgFrame The frame to position.
		 * @param pkgName The package name.
		 * @param bounds The bounds of the previous package.
		 * @param boundsPkgName The name of the previous package.
		 */
		public MyPkgPositioner(Frame pkgFrame, String pkgName, Rectangle bounds, String boundsPkgName) {
			this.pkgFrame = pkgFrame;
			this.pkgName = pkgName;
			this.x = bounds.x;
			this.y = bounds.y;
			this.width = bounds.width;
			this.height = bounds.height;
			this.boundsPkgName = boundsPkgName;
			attempts = 1;
		}
		
		/**
		 * Set the number of attempts to position the window.
		 * @param attempts Number of attempts (100ms each).
		 */
		public void setAttempts(int attempts) {
			this.attempts = attempts;
		}
		
		public void run() {
			if(pkgFrame==null || pkgName==null) return;
			final Frame f = pkgFrame;
			final Point newLocation;
			if(jCheckBox4.isSelected()) {
				newLocation = myCascadePackage(pkgName, new Point(x, y), boundsPkgName); 
			} else if(jCheckBox3.isSelected()){
				newLocation = new Point(x, y); 
			} else {
				newLocation = null;
			}
			final Dimension newSize;
			if(jCheckBox5.isSelected()) {
				newSize = new Dimension(width, height);
			} else {
				newSize = null;
			}
			
			// set package window position
			for(int i=0; i<attempts; i++) {
				SwingUtilities.invokeLater(new Runnable() {
					public void run() {
						if(newLocation!=null) f.setLocation(newLocation);
						if(newSize!=null) f.setSize(newSize);
						f.validate();
					}
				});
				try { Thread.sleep(100); } catch (InterruptedException e) {}
				
				// then check if window has the new position and/or size
				// (if not: repeat <attempts> times)
				if(
					(newLocation==null || pkgFrame.getLocation().equals(newLocation))
					&& (newSize==null || pkgFrame.getSize().equals(newSize))
				) break;
			}
		}
	}
	
	/**
	 * A class representing a file,
	 * having alternative toString and equals methods.
	 * @author Manuel Haim
	 */
	class MyFile implements Comparable {
		private File file;
		public MyFile(File file) {
			this.file = file;
		}
		public File getFile() {
			return file;
		}
		public void setFile(File file) {
			this.file = file;
		}
		public String toString() {
			if(!file.exists()) return file.getName() + "*";
			return file.getName();
		}
		public boolean equals(Object o) {
			if(o instanceof MyFile) {
				if(((MyFile) o).file.getName().equals(this.file.getName())) return true;
			}
			return false;
		}
		public int compareTo(Object o) {
			if(o instanceof MyFile) {
				return this.toString().compareTo( ((MyFile) o).toString() );
			}
			return -1;
		}
	}

	/**
	 * An array based list model.
	 * @author Manuel Haim
	 */
	class MyListModel extends javax.swing.AbstractListModel {

		private static final long serialVersionUID = 1L;

		private Object[] data;
		
		public MyListModel(Object[] data) {
			this.data = data;
		}
		
		public Object getElementAt(int index) {
			return data[index];
		}

		public int getSize() {
			return data.length;
		}
	}
	
	/**
	 * Get the import directory.
	 * @return The import directory (if valid), null otherwise.
	 */
	private File myGetImportDir() {
		String importDirPath = jTextField.getText();
		File importDir = new File(importDirPath);
		if(importDirPath.equals("") || !importDir.isDirectory()) return null;
		return importDir;
	}
	
	/**
	 * Check if there is a jar file to unpack from.
	 * Please note: To check for subdirs, use {@link #myGetUnpackSubdir(File)}. 
	 * @param file The target project within the workspace.
	 * @return true if a project-specific jar file exists, false otherwise.
	 */
	private boolean myUnpackJarfileExists(File file) {
		File importDir = myGetImportDir();
		if(importDir==null) return false;
		// find jar file (note: a file ending in .JaR is a jar file, too)
		File[] rootFiles = importDir.listFiles();
		for(int i=0; i<rootFiles.length; i++) {
			if(rootFiles[i].isFile()
			&& rootFiles[i].getName().startsWith(file.getName())
			&& rootFiles[i].getName().length() == file.getName().length() + ".jar".length()
			&& rootFiles[i].getName().toLowerCase().endsWith(".jar")) {
				return true;
			}
		}
		return false;
	}
	
	/**
	 * Unpack a project from the import directory (only if non-existing).
	 * @param file The target project within the workspace.
	 */
	private void myUnpack(File file) {
		if(file.isDirectory()) return;
		
		file.mkdirs();
		if(!file.isDirectory()) return;
		File importDir = myGetImportDir();
		if(importDir==null) return;
		// find jar file (note: a file ending in .JaR is a jar file, too)
		File[] rootFiles = importDir.listFiles();
		for(int i=0; i<rootFiles.length; i++) {
			if(rootFiles[i].isFile()
			&& rootFiles[i].getName().startsWith(file.getName())
			&& rootFiles[i].getName().length() == file.getName().length() + ".jar".length()
			&& rootFiles[i].getName().toLowerCase().endsWith(".jar")) {
				//unpack
				FileHelpers.unpack(rootFiles[i], file, false);
			}
		}
		
		File importSubdir = myGetUnpackSubdir(file);
		if(importSubdir==null) return;
		File[] subdirFiles = importSubdir.listFiles();
		for(int i=0; i<subdirFiles.length; i++) {
			myDoUnpackCopy(subdirFiles[i], file);
		}
	}
	
	/**
	 * Unpack all projects from the import directory.
	 */
	private void myUnpackAll() {
		int option = JOptionPane.showConfirmDialog(null, "Overwrite existing projects?", "Import all projects", JOptionPane.YES_NO_CANCEL_OPTION);
		if(option==JOptionPane.CANCEL_OPTION || option==JOptionPane.CLOSED_OPTION) return;
		boolean overwrite = (option==JOptionPane.YES_OPTION);
		
		Set createdProjects = new TreeSet();
		
		boolean checkSubdirs = !myImportDirEqualsWorkspaceDir();
		
		// iterate all files in import dir
		File importDir = myGetImportDir();
		if(importDir==null) return;
		File[] rootFiles = importDir.listFiles();
		for(int i=0; i<rootFiles.length; i++) {
			// check for jar file
			if(rootFiles[i].isFile()
			&& rootFiles[i].getName().toLowerCase().endsWith(".jar")) {
				//unpack jar file
				File file = new File(
					FileHelpers.addFileSeparator(lastRefreshedWorkspace)
					+ rootFiles[i].getName().substring(0, rootFiles[i].getName().length() - ".jar".length())
				);
				
				//check if target project already exists
				if(file.isDirectory() && !createdProjects.contains(file.getName())) {
					if(!overwrite) continue; //next jar file or subdir
					boolean deleted = false;
					if(bluej!=null) deleted = myFindDelete(file);
					if(!deleted) deleted = FileHelpers.deleteDir(file);
					if(!deleted) continue;
				}
				createdProjects.add(file.getName());
				try {
						FileHelpers.unpack(rootFiles[i], file, false);
				} catch (Exception e) {}
			}
			
			// check for directory
			if(checkSubdirs && rootFiles[i].isDirectory()) {
				//unpack/copy subdirectory files
				File importSubdir = rootFiles[i];
				File file = new File(
					FileHelpers.addFileSeparator(lastRefreshedWorkspace)
					+ importSubdir.getName()
				);
				
				//check if target project already exists
				if(file.isDirectory() && !createdProjects.contains(file.getName())) {
					if(!overwrite) continue; //next jar file or subdir
					boolean deleted = false;
					if(bluej!=null) deleted = myFindDelete(file);
					if(!deleted) deleted = FileHelpers.deleteDir(file);
					if(!deleted) continue;
				}
				createdProjects.add(file.getName());
				file.mkdirs();
				File[] subdirFiles = importSubdir.listFiles();
				try {
					for(int j=0; j<subdirFiles.length; j++) {
						myDoUnpackCopy(subdirFiles[j], file);
					}
				} catch (Exception e) {}
			}
		}
	}

	/**
	 * Unpack or copy a file from the project-specific import subdirectory.
	 * @param subdirFile A file from the import subdirectory.
	 * @param file The target project within the workspace.
	 */
	private void myDoUnpackCopy(File subdirFile, File file) {
		if(subdirFile.isDirectory()) {
			if(jCheckBox.isSelected()) FileHelpers.copyDirectory(
				subdirFile,
				new File(FileHelpers.addFileSeparator(file.getAbsolutePath()) + subdirFile.getName()),
				false
			);
		}
		if(subdirFile.isFile()) {
			if(jCheckBox2.isSelected() && subdirFile.getName().toLowerCase().endsWith(".jar")) {
				// unzip jar file
				FileHelpers.unpack(subdirFile, file, false);
			} else if(jCheckBox.isSelected()) {
				// copy
				FileHelpers.copyFile(subdirFile,
					new File(
						FileHelpers.addFileSeparator(file.getAbsolutePath()) + subdirFile.getName()
					),
					false
				);
			}
		}
	}
	
	
	/**
	 * Check if import dir is equal to workspace dir.
	 * @return true if equal (or import dir non-existing), false otherwise.
	 */
	private boolean myImportDirEqualsWorkspaceDir() {
		File importDir = myGetImportDir();
		if(importDir==null) return true;
		
		File workspaceDir = new File(lastRefreshedWorkspace);
		
		// If the import dir equals the workspace dir, all subdirs are projects!
		if(importDir.getAbsolutePath().equals(workspaceDir.getAbsolutePath())) return true;
		return false;
	}
	
	/**
	 * Get the import subdirectory for a project.
	 * @param file The target project within the workspace.
	 * @return The import subdirectory or null if (not existing or within workspace).
	 */
	private File myGetUnpackSubdir(File file) {
		if(myImportDirEqualsWorkspaceDir()) return null;
		
		File importDir = myGetImportDir();
		if(importDir==null) return null;
		String importDirPath = FileHelpers.addFileSeparator(importDir.getAbsolutePath());
		
		
		File importFile = (new File(importDirPath + file.getName())).getAbsoluteFile();
		if(!importFile.exists() || !importFile.isDirectory()) return null;
		return importFile;
	}
	
	
	/**
	 * Unpack the selected project from the import directory
	 * (only if non-existing).
	 */
	private void myUnpack() {
		File file = ((MyFile) jList.getSelectedValue()).getFile();
		myUnpack(file);
	}
	
	/**
	 * Open the selected project as BlueJ project.
	 */
	private void myOpen() {
		File file = ((MyFile) jList.getSelectedValue()).getFile();
		
		// create BlueJ project file (if missing)
		FileHelpers.createBluejpkg(file);
		
		if(bluej!=null) {
			bluej.openProject(file);
		}
	}

	/**
	 * Refresh the project list by reloading the workspace directory.
	 */
	private void myRefreshList() {
		// the list will be sorted
		SortedSet s = new TreeSet();

		// the Workspace directory
		if(jComboBox.getSelectedItem()==null) return;
		lastRefreshedWorkspace = jComboBox.getSelectedItem().toString();
		File file = (new File(lastRefreshedWorkspace)).getAbsoluteFile();
		if(file.isDirectory()) {
			File[] childArr = file.listFiles();
			for (int i = 0; i < childArr.length; i++) {
				if(childArr[i].isDirectory()) {
					// treat all directories as project
					s.add(new MyFile(childArr[i]));
				}
			}
		}

		// the Import directory
		String importDir = jTextField.getText();
		if(!importDir.equals("")) {
			file = (new File(importDir)).getAbsoluteFile();
			String toDir = FileHelpers.addFileSeparator(lastRefreshedWorkspace);
			if(file.isDirectory()) {
				File[] childArr = file.listFiles();
				for (int i = 0; i < childArr.length; i++) {
					if(childArr[i].isFile()) {
						if(childArr[i].getName().toLowerCase().endsWith(".jar")) {
							// if jar file found: set project path
							s.add(new MyFile( (new File(toDir + childArr[i].getName().substring(0, childArr[i].getName().length() - ".jar".length()))).getAbsoluteFile() ));
						}
					}
					if(childArr[i].isDirectory()) {
						// if directory found: set project path
						s.add(new MyFile( (new File(toDir + childArr[i].getName())).getAbsoluteFile() ));
					}
				}
			} else {
			}
		}

		jList.setListData(s.toArray(new MyFile[0]));
	}

	/**
	 * Pack the selected project to a jar file.
	 */
	private void myPack() {
		File dir = ((MyFile) jList.getSelectedValue()).getFile();
		if(!dir.isDirectory()) return;
		String toDir = FileHelpers.addFileSeparator(jComboBox.getSelectedItem().toString());
		
		if(!toDir.endsWith(File.separator)) toDir += File.separator;
		File destFile = new File(toDir + dir.getName() + ".jar");
		if(destFile.exists()) {
			if(JOptionPane.showConfirmDialog(null, "File exists, overwrite?\n" + destFile.toString(), "Pack project", JOptionPane.OK_CANCEL_OPTION)
					== JOptionPane.OK_OPTION)
				destFile.delete();
		}
		FileHelpers.pack(dir, destFile, jCheckBox1.isSelected());
	}

	/**
	 * Pack all projects to jar files.
	 */
	private void myPackAll() {
		if(jList.getModel().getSize()==0) return;
		int result = JOptionPane.showConfirmDialog(null, "Shall existing files be overwritten?", "Pack all projects", JOptionPane.YES_NO_CANCEL_OPTION);
		if(result == JOptionPane.CANCEL_OPTION || result == JOptionPane.CLOSED_OPTION) return;
		boolean overwrite = (result	== JOptionPane.YES_OPTION);
		for(int i=0; i<jList.getModel().getSize(); i++) {
			try {
				File dir = ((MyFile) jList.getModel().getElementAt(i)).getFile();
				String toDir = FileHelpers.addFileSeparator(jComboBox.getSelectedItem().toString());

				if(!toDir.endsWith(File.separator)) toDir += File.separator;
				File destFile = new File(toDir + dir.getName() + ".jar");
				if(destFile.exists() && overwrite) destFile.delete();
				FileHelpers.pack(dir, destFile, jCheckBox1.isSelected());
			} catch (Exception e) {}
		}
	}
	
	/**
	 * This method initializes jButton6	
	 * 	
	 * @return javax.swing.JButton	
	 */
	private JButton getJButton6() {
		if (jButton6 == null) {
			jButton6 = new JButton();
			jButton6.setText("Select workspace...");
			jButton6.setToolTipText("Open a file requester to select a workspace (dir with projects).");
			jButton6.addActionListener(new java.awt.event.ActionListener() {
				public void actionPerformed(java.awt.event.ActionEvent e) {
					if(fileChooser==null) fileChooser = new JFileChooser();
					
					fileChooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
					fileChooser.setMultiSelectionEnabled(false);
					fileChooser.setDialogTitle("Open workspace...");
					Object comboSelected = jComboBox.getSelectedItem();
					if(comboSelected!=null) {
						File dir = (new File(comboSelected.toString())).getAbsoluteFile();
						if(dir.isDirectory()) fileChooser.setSelectedFile(dir);
					}
					
					if (fileChooser.showOpenDialog(null) == JFileChooser.APPROVE_OPTION) {
						if(fileChooser.getSelectedFile().isDirectory()) {
							Object workspace = fileChooser.getSelectedFile().getPath();
							jComboBox.addItem(workspace);
							jComboBox.setSelectedItem(workspace);
							comboBoxChange(workspace); // seems this has to be called manually here
						}
					}
				}
			});
		}
		return jButton6;
	}

	/**
	 * This method initializes jButton7	
	 * 	
	 * @return javax.swing.JButton	
	 */
	private JButton getJButton7() {
		if (jButton7 == null) {
			jButton7 = new JButton();
			jButton7.setText("Set import dir...");
			jButton7.setToolTipText("Open a file requester to select an import dir for Unpack (workspace-specific, must contain Java Archive files).");
			jButton7.addActionListener(new java.awt.event.ActionListener() {
				public void actionPerformed(java.awt.event.ActionEvent e) {
					if(fileChooser==null) fileChooser = new JFileChooser();
					
					fileChooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
					fileChooser.setMultiSelectionEnabled(false);
					fileChooser.setDialogTitle("Select import directory...");
					
					// Set fileChooser dir to current import dir (if available)
					// or project workspace (if available)
					String currentImportDir = jTextField.getText();
					Object comboSelected = lastRefreshedWorkspace;
					if(!currentImportDir.equals("")
					&& (new File(currentImportDir)).isDirectory())
						comboSelected = currentImportDir;
					if(comboSelected!=null) {
						File dir = (new File(comboSelected.toString())).getAbsoluteFile();
						if(dir.isDirectory()) fileChooser.setSelectedFile(dir);
					}
					
					if (fileChooser.showOpenDialog(null) == JFileChooser.APPROVE_OPTION) {
						if(fileChooser.getSelectedFile().isDirectory()) {
							jTextField.setText(fileChooser.getSelectedFile().getPath());
							myRefreshList();
						}
					}
				}
			});
		}
		return jButton7;
	}

	/**
	 * This method initializes jPanel4	
	 * 	
	 * @return javax.swing.JPanel	
	 */
	private JPanel getJPanel4() {
		if (jPanel4 == null) {
			jPanel4 = new JPanel();
			jPanel4.setLayout(new GridLayout());
			jPanel4.add(getJTextField());
		}
		return jPanel4;
	}


	/**
	 * This method initializes jButton8	
	 * 	
	 * @return javax.swing.JButton	
	 */
	private JButton getJButton8() {
		if (jButton8 == null) {
			jButton8 = new JButton();
			jButton8.setText("Browse");
			jButton8.setToolTipText("Open directory of selected project by the default directory browser.");
			jButton8.addActionListener(new java.awt.event.ActionListener() {
				public void actionPerformed(java.awt.event.ActionEvent e) {
					File file = null;
					if(jList.getSelectedIndex()==-1) {
						if(lastRefreshedWorkspace!=null)
							file = new File(lastRefreshedWorkspace);
					} else {
						file = ((MyFile) jList.getSelectedValue()).getFile();
					}
					if(file!=null && file.isDirectory()) FileHelpers.openDocument(file);
				}
			});
		}
		return jButton8;
	}

	/**
	 * This method initializes jPanel7	
	 * 	
	 * @return javax.swing.JPanel	
	 */
	private JPanel getJPanel7() {
		if (jPanel7 == null) {
			jPanel7 = new JPanel();
			jPanel7.setLayout(new BoxLayout(jPanel7, BoxLayout.PAGE_AXIS));
			jPanel7.add(getJPanel11());
			jPanel7.add(getJPanel10());
			jPanel7.add(getJPanel12());
			jPanel7.add(getJPanel19());
		}
		return jPanel7;
	}

	/**
	 * This method initializes jButton9	
	 * 	
	 * @return javax.swing.JButton	
	 */
	private JButton getJButton9() {
		if (jButton9 == null) {
			jButton9 = new JButton();
			jButton9.setText("Unpack all");
			jButton9.setToolTipText("Import/Unpack all projects from import dir (no open, ask before overwrite).");
			jButton9.addActionListener(new java.awt.event.ActionListener() {
				public void actionPerformed(java.awt.event.ActionEvent e) {
					try {
						projectChangeMutex.acquire();
						projectChange.acquire();
						myUnpackAll();
					} catch(InterruptedException e1) {
					} finally {
						projectChange.release();
						projectChangeMutex.release();
					}
					myRefreshList();
				}
			});
		}
		return jButton9;
	}

	/**
	 * Switch from one project to another and do window positioning.
	 * The selected project is unpacked and opened after closing the others.
	 */
	private void mySwitch() {
		myUnpack();
		jList.repaint();
		myCloseAll();
		myOpen();
	}

	/**
	 * This method initializes jTabbedPane	
	 * 	
	 * @return javax.swing.JTabbedPane	
	 */
	private JTabbedPane getJTabbedPane() {
		if (jTabbedPane == null) {
			jTabbedPane = new JTabbedPane();
			jTabbedPane.addTab("Projects", null, getJPanel3(), "All projects in workspace.");
			jTabbedPane.addTab("Options", null, getJPanel7(), "Set workspace-specific options.");
		}
		return jTabbedPane;
	}

	/**
	 * This method initializes jPanel1	
	 * 	
	 * @return javax.swing.JPanel	
	 */
	private JPanel getJPanel1() {
		if (jPanel1 == null) {
			GridBagConstraints gridBagConstraints = new GridBagConstraints();
			gridBagConstraints.fill = GridBagConstraints.BOTH;
			gridBagConstraints.gridy = 0;
			gridBagConstraints.weightx = 1.0;
			gridBagConstraints.weighty = 1.0;
			gridBagConstraints.gridx = 0;
			jPanel1 = new JPanel();
			jPanel1.setLayout(new GridBagLayout());
			jPanel1.add(getJTabbedPane(), gridBagConstraints);
		}
		return jPanel1;
	}

	/**
	 * This method initializes jPanel8	
	 * 	
	 * @return javax.swing.JPanel	
	 */
	private JPanel getJPanel8() {
		if (jPanel8 == null) {
			GridBagConstraints gridBagConstraints1 = new GridBagConstraints();
			gridBagConstraints1.fill = GridBagConstraints.BOTH;
			gridBagConstraints1.gridy = 0;
			gridBagConstraints1.weightx = 1.0;
			gridBagConstraints1.weighty = 1.0;
			gridBagConstraints1.gridx = 0;
			jPanel8 = new JPanel();
			jPanel8.setLayout(new GridBagLayout());
			jPanel8.add(getJScrollPane(), gridBagConstraints1);
		}
		return jPanel8;
	}

	/**
	 * This method initializes jPanel9	
	 * 	
	 * @return javax.swing.JPanel	
	 */
	private JPanel getJPanel9() {
		if (jPanel9 == null) {
			jPanel9 = new JPanel();
			jPanel9.setLayout(new GridLayout(0,1));
			jPanel9.add(getJCheckBox2());
			jPanel9.add(getJCheckBox());
		}
		return jPanel9;
	}

	/**
	 * This method initializes jPanel10	
	 * 	
	 * @return javax.swing.JPanel	
	 */
	private JPanel getJPanel10() {
		if (jPanel10 == null) {
			jPanel10 = new JPanel();
			jPanel10.setLayout(new BoxLayout(jPanel10, BoxLayout.PAGE_AXIS));
			jPanel10.setBorder(BorderFactory.createTitledBorder("Export / Pack options"));
			jPanel10.add(getJPanel13());
		}
		return jPanel10;
	}

	/**
	 * This method initializes jCheckBox	
	 * 	
	 * @return javax.swing.JCheckBox	
	 */
	private JCheckBox getJCheckBox() {
		if (jCheckBox == null) {
			jCheckBox = new JCheckBox();
			jCheckBox.setText("copy files from subdirs");
			jCheckBox.setToolTipText("If checked, remaining files will be copied from project-specific subdirs.");
		}
		return jCheckBox;
	}

	/**
	 * This method initializes jPanel11	
	 * 	
	 * @return javax.swing.JPanel	
	 */
	private JPanel getJPanel11() {
		if (jPanel11 == null) {
			jPanel11 = new JPanel();
			jPanel11.setLayout(new BoxLayout(jPanel11, BoxLayout.PAGE_AXIS));
			jPanel11.setBorder(BorderFactory.createTitledBorder("Unpack options"));
			jPanel11.add(getJPanel());
			jPanel11.add(getJPanel4());
			jPanel11.add(getJPanel9());
		}
		return jPanel11;
	}

	/**
	 * This method initializes jPanel12	
	 * 	
	 * @return javax.swing.JPanel	
	 */
	private JPanel getJPanel12() {
		if (jPanel12 == null) {
			jPanel12 = new JPanel();
			jPanel12.setLayout(new BoxLayout(jPanel12, BoxLayout.PAGE_AXIS));
			jPanel12.setBorder(BorderFactory.createTitledBorder("BlueJ window options"));
			jPanel12.add(getJPanel20(), null);
		}
		return jPanel12;
	}

	/**
	 * This method initializes jPanel13	
	 * 	
	 * @return javax.swing.JPanel	
	 */
	private JPanel getJPanel13() {
		if (jPanel13 == null) {
			jPanel13 = new JPanel();
			jPanel13.setLayout(new GridLayout());
			jPanel13.add(getJCheckBox1());
		}
		return jPanel13;
	}

	/**
	 * This method initializes jCheckBox1	
	 * 	
	 * @return javax.swing.JCheckBox	
	 */
	private JCheckBox getJCheckBox1() {
		if (jCheckBox1 == null) {
			jCheckBox1 = new JCheckBox();
			jCheckBox1.setText("exclude javadoc");
			jCheckBox1.setToolTipText("If checked, the \"doc\" directory will not appear in the output archive.");
		}
		return jCheckBox1;
	}

	/**
	 * This method initializes jButton10	
	 * 	
	 * @return javax.swing.JButton	
	 */
	private JButton getJButton10() {
		if (jButton10 == null) {
			jButton10 = new JButton();
			jButton10.setText("New");
			jButton10.setToolTipText("Create new project (ask for name).");
			jButton10.addActionListener(new java.awt.event.ActionListener() {
				public void actionPerformed(java.awt.event.ActionEvent e) {
					String name = JOptionPane.showInputDialog(null, "Enter project name:", "New project", JOptionPane.QUESTION_MESSAGE);
					if(name!=null) {
						File file = new File(FileHelpers.addFileSeparator(lastRefreshedWorkspace) + name);
						boolean created = false;
						if(!name.equals("") && file.getName().equals(name) && !file.exists()) {
							if(bluej==null) {
								created = file.mkdir();
								myRefreshList();
							} else {
								try {
									projectChangeMutex.acquire();
									projectChange.acquire();
									myCloseAll();
									bluej.newProject(file);
									myRefreshList();
									created = (myFindOpen()!=null);
								} catch (InterruptedException e1) {
								} finally {
									projectChange.release();
									projectChangeMutex.release();
								}
							}
						}
						if(!created) JOptionPane.showMessageDialog(null, "Cannot create project:\n" + file.getAbsolutePath());
					}
				}
			});
		}
		return jButton10;
	}

	/**
	 * This method initializes jButton11	
	 * 	
	 * @return javax.swing.JButton	
	 */
	private JButton getJButton11() {
		if (jButton11 == null) {
			jButton11 = new JButton();
			jButton11.setText("Close all");
			jButton11.setToolTipText("Close all currently open projects.");
			jButton11.addActionListener(new java.awt.event.ActionListener() {
				public void actionPerformed(java.awt.event.ActionEvent e) {
					jList.clearSelection();
					if(bluej!=null) myCloseAll();
				}
			});
		}
		return jButton11;
	}

	/**
	 * Try to find the currently open project in project list ({@link #bluej} required).
	 * If it is in list, the list item will be selected.
	 * @return The project (if found), null otherwise.
	 */
	private BProject myFindOpen() {
		jList.clearSelection();
		BPackage currentPackage = bluej.getCurrentPackage();
		if(currentPackage!=null) try {
			BProject lastOpened = currentPackage.getProject();
			File lastOpenedFile = lastOpened.getDir();
			for(int i=0; i<jList.getModel().getSize(); i++) {
				File listFile = ((MyFile) jList.getModel().getElementAt(i)).getFile();
				if(listFile!=null
				&& listFile.getAbsolutePath().equals(lastOpenedFile.getAbsolutePath())) {
					jList.setSelectedIndex(i);
					return lastOpened;
				}
			}
		} catch (ProjectNotOpenException e) {}
		return null;
	}

	/**
	 * Find an open project, close and delete it ({@link #bluej} required).
	 * @param file The project dir.
	 * @return true if found and deleted, false otherwise.
	 */
	private boolean myFindDelete(File file) {
		boolean deleted = false;
		boolean projectFound = myFindClose(file);
		if(projectFound) {
			// Wait for BlueJ to release file handles (try 10 times):
			try {
				for(int j=0; j<10 && deleted==false; j++) {
					deleted = FileHelpers.deleteDir(file);
					Thread.sleep(100);
				}
			} catch (InterruptedException e1) {}
		}
		return deleted;
	}

	/**
	 * Find an open project and close it ({@link #bluej} required).
	 * @param file The project dir.
	 * @return true if project was found and closed, false otherwise.
	 */
	private boolean myFindClose(File file) {
		boolean projectFound = false;
		BProject[] openProjects = bluej.getOpenProjects();
		for(int k=0; k<openProjects.length; k++) {
			try {
				if(openProjects[k].getDir().getAbsolutePath().equals(file.getAbsolutePath())) {
					projectFound = true;
					openProjects[k].close();
				}
			} catch (ProjectNotOpenException e1) {}
		}
		return projectFound;
	}

	/**
	 * Close all currently open projects ({@link #bluej} required).
	 */
	private void myCloseAll() {
		BProject[] openProjects = bluej.getOpenProjects();
		for(int k=0; k<openProjects.length; k++) {
			try {
				openProjects[k].close();
			} catch (ProjectNotOpenException e1) {}
		}
	}

	/**
	 * This method initializes jButton12	
	 * 	
	 * @return javax.swing.JButton	
	 */
	private JButton getJButton12() {
		if (jButton12 == null) {
			jButton12 = new JButton();
			jButton12.setText("Import");
			jButton12.setToolTipText("Import a Java Archive file into workspace (open file requester, ask for name).");
			jButton12.addActionListener(new java.awt.event.ActionListener() {
				public void actionPerformed(java.awt.event.ActionEvent e) {
					if(fileChooser2==null) {
						fileChooser2 = new JFileChooser();
						fileChooser2.setFileSelectionMode(JFileChooser.FILES_ONLY);
						fileChooser2.setFileFilter(new FileFilter() {
							public boolean accept(File f) {
								if(f.isDirectory() || f.getName().toLowerCase().endsWith(".jar")) return true;
								return false;
							}

							public String getDescription() {
								return "Java Archives (*.jar)";
							}
							
						});
					}

					fileChooser2.setMultiSelectionEnabled(false);
					fileChooser2.setDialogTitle("Import a project...");

					if (fileChooser2.showOpenDialog(null) == JFileChooser.APPROVE_OPTION) {
						File jarFile = fileChooser2.getSelectedFile();
						if(jarFile.exists() && !jarFile.isDirectory() && jarFile.getName().toLowerCase().endsWith(".jar")) {
							String name = jarFile.getName().substring(0,jarFile.getName().length()-".jar".length());
							name = JOptionPane.showInputDialog(null, "Enter project name for Java Archive import:\n" + jarFile.getName(), name); 
							
							if(name!=null) {
								File file = new File(FileHelpers.addFileSeparator(lastRefreshedWorkspace) + name);
								
								if(!name.equals("") && file.isDirectory()) {
									if(JOptionPane.showConfirmDialog(null, "Project exists, overwrite?\n" + file.getAbsolutePath(), "Import Java Archive file", JOptionPane.OK_CANCEL_OPTION)
										==JOptionPane.OK_OPTION) {
										boolean deleted = myFindDelete(file);
										if(!deleted) FileHelpers.deleteDir(file);
									} else return;
								}

								boolean created = false;
								if(!name.equals("") && file.getName().equals(name) && !file.exists()) {
									FileHelpers.unpack(jarFile, file, false);
									created = file.isDirectory(); // TODO Let FileHelpers.unpack return boolean value.
									if(bluej==null) {
										myRefreshList();
									} else {
										try {
											projectChangeMutex.acquire();
											projectChange.acquire();
											myCloseAll();
											
											// create BlueJ project file (if missing)
											FileHelpers.createBluejpkg(file);
											
											bluej.openProject(file);
											myRefreshList();
											created = (myFindOpen()!=null);
										} catch (InterruptedException e1) {
										} finally {
											projectChange.release();
											projectChangeMutex.release();
										}
									}
								}
								if(!created) JOptionPane.showMessageDialog(null, "Cannot create project:\n" + file.getAbsolutePath());
							}
							
							
							
							
							
						}
					}
					
				}
			});
		}
		return jButton12;
	}

	/**
	 * This method initializes jButton13	
	 * 	
	 * @return javax.swing.JButton	
	 */
	private JButton getJButton13() {
		if (jButton13 == null) {
			jButton13 = new JButton();
			jButton13.setText("Export");
			jButton13.setToolTipText("Export selected project as Java Archive file (open file requester).");
			jButton13.addActionListener(new java.awt.event.ActionListener() {
				public void actionPerformed(java.awt.event.ActionEvent e) {
					if(jList.getSelectedIndex()==-1) return;
					if(fileChooser2==null) {
						fileChooser2 = new JFileChooser();
						fileChooser2.setFileSelectionMode(JFileChooser.FILES_ONLY);
						fileChooser2.setFileFilter(new FileFilter() {
							public boolean accept(File f) {
								if(f.isDirectory() || f.getName().toLowerCase().endsWith(".jar")) return true;
								return false;
							}

							public String getDescription() {
								return "Java Archives (*.jar)";
							}
							
						});
					}
					
					fileChooser2.setMultiSelectionEnabled(false);
					fileChooser2.setDialogTitle("Export selected project...");
					File dir = ((MyFile) jList.getSelectedValue()).getFile();
					fileChooser2.setSelectedFile( new File(dir.getName() + ".jar") );
					
					if (fileChooser2.showSaveDialog(null) == JFileChooser.APPROVE_OPTION) {
						boolean saveFile = true;
						File file = fileChooser2.getSelectedFile();
						if(file.exists()) {
							//Overwrite?
							saveFile = JOptionPane.showConfirmDialog(null, "Overwrite existing Java Archive file?\n" + file.getAbsolutePath(), "Save Java Archive file", JOptionPane.OK_CANCEL_OPTION)==JOptionPane.OK_OPTION;
							if(saveFile) saveFile = file.delete();
						}
						if(saveFile) {
							FileHelpers.pack(dir, file, jCheckBox1.isSelected());
							if(!file.exists()) JOptionPane.showMessageDialog(null, "Error writing Java Archive file:\n" + file.getAbsolutePath());
						}
						
					}
				}
			});
		}
		return jButton13;
	}

	/**
	 * This method initializes jTabbedPane1	
	 * 	
	 * @return javax.swing.JTabbedPane	
	 */
	private JTabbedPane getJTabbedPane1() {
		if (jTabbedPane1 == null) {
			jTabbedPane1 = new JTabbedPane();
			//jTabbedPane1.setSize(new Dimension(153, 104));
			//jTabbedPane1.setTabPlacement(JTabbedPane.BOTTOM);
			
			jTabbedPane1.addTab("Basic", null, getJPanel14(), "Basic project operations.");
			jTabbedPane1.addTab("Advanced", null, getJPanel15(), "Advanced project operations.");
		}
		return jTabbedPane1;
	}

	/**
	 * This method initializes jPanel14	
	 * 	
	 * @return javax.swing.JPanel	
	 */
	private JPanel getJPanel14() {
		if (jPanel14 == null) {
			jPanel14 = new JPanel();
			jPanel14.setLayout(new BoxLayout(jPanel14, BoxLayout.PAGE_AXIS));
			jPanel14.add(getJPanel2());
			jPanel14.add(getJPanel18());
		}
		return jPanel14;
	}

	/**
	 * This method initializes jPanel15	
	 * 	
	 * @return javax.swing.JPanel	
	 */
	private JPanel getJPanel15() {
		if (jPanel15 == null) {
			jPanel15 = new JPanel();
			jPanel15.setLayout(new BoxLayout(jPanel15, BoxLayout.PAGE_AXIS));
			jPanel15.add(getJPanel16());
			jPanel15.add(getJPanel17());
		}
		return jPanel15;
	}

	/**
	 * This method initializes jPanel16	
	 * 	
	 * @return javax.swing.JPanel	
	 */
	private JPanel getJPanel16() {
		if (jPanel16 == null) {
			jPanel16 = new JPanel();
			jPanel16.setLayout(new GridLayout(0,2));
			jPanel16.add(getJButton2());
			jPanel16.add(getJButton16());
			jPanel16.add(getJButton14());
			jPanel16.add(getJButton17());
			jPanel16.add(getJButton());
			jPanel16.add(getJButton9());
			jPanel16.add(getJButton4());
			jPanel16.add(getJButton5());
		}
		return jPanel16;
	}

	/**
	 * This method initializes jPanel17	
	 * 	
	 * @return javax.swing.JPanel	
	 */
	private JPanel getJPanel17() {
		if (jPanel17 == null) {
			jPanel17 = new JPanel();
			jPanel17.setLayout(new GridBagLayout());
		}
		return jPanel17;
	}

	/**
	 * This method initializes jPanel18	
	 * 	
	 * @return javax.swing.JPanel	
	 */
	private JPanel getJPanel18() {
		if (jPanel18 == null) {
			jPanel18 = new JPanel();
			jPanel18.setLayout(new GridBagLayout());
		}
		return jPanel18;
	}

	/**
	 * This method initializes jButton14	
	 * 	
	 * @return javax.swing.JButton	
	 */
	private JButton getJButton14() {
		if (jButton14 == null) {
			jButton14 = new JButton();
			jButton14.setText("Rename");
			jButton14.setToolTipText("Rename selected project (ask for name).");
			jButton14.addActionListener(new java.awt.event.ActionListener() {
				public void actionPerformed(java.awt.event.ActionEvent e) {
					if(jList.getSelectedIndex()==-1) return;
					File file = ((MyFile) jList.getSelectedValue()).getFile();
					boolean renamed = false;
					boolean projectFound = false;
					File destFile = null;
					if(file.isDirectory()) {
						String name = JOptionPane.showInputDialog(null, "Enter new project name:", file.getName());
						if(name!=null && !name.equals("")) {
							if(bluej!=null) try {
								projectChangeMutex.acquire();
								projectChange.acquire();
								projectFound = myFindClose(file);
							} catch (InterruptedException e1) {
							} finally {
								projectChange.release();
								projectChangeMutex.release();
							}
							String newPath = file.getAbsolutePath().substring(
								0, file.getAbsolutePath().length() - file.getName().length()
							) + name;
							destFile = new File(newPath);
							if(destFile.getName().equals(name)) {
								if(projectFound) {
									// Wait for BlueJ to release file handles (try 10 times):
									try {
										for(int j=0; j<10 && renamed==false; j++) {
											renamed = file.renameTo(destFile);
											Thread.sleep(100);
										}
									} catch (InterruptedException e1) {}
								}
								if(!renamed) renamed = file.renameTo(destFile);
							}
						} else return;
					}
					if(renamed) {
						myRefreshList();
						if(bluej!=null && projectFound) try {
							projectChangeMutex.acquire();
							projectChange.acquire();
							bluej.openProject(destFile);
							myFindOpen();
						} catch (InterruptedException e1) {
						} finally {
							projectChange.release();
							projectChangeMutex.release();
						}
					}
					if(!renamed) JOptionPane.showMessageDialog(null, "Could not rename project:\n" + file.getAbsolutePath());
				}
			});
		}
		return jButton14;
	}

	/**
	 * This method initializes jButton15	
	 * 	
	 * @return javax.swing.JButton	
	 */
	private JButton getJButton15() {
		if (jButton15 == null) {
			jButton15 = new JButton();
			jButton15.setText("Open");
			jButton15.setToolTipText("Open the selected project (*unpack if non-existing).");
			jButton15.addActionListener(new java.awt.event.ActionListener() {
				public void actionPerformed(java.awt.event.ActionEvent e) {
					if(jList.getSelectedIndex()==-1) return;
					try {
						projectChangeMutex.acquire();
						projectChange.acquire();
						mySwitch();
					} catch (InterruptedException e1) {
					} finally {
						projectChange.release();
						projectChangeMutex.release();
					}
				}
			});
		}
		return jButton15;
	}

	/**
	 * This method initializes jCheckBox2	
	 * 	
	 * @return javax.swing.JCheckBox	
	 */
	private JCheckBox getJCheckBox2() {
		if (jCheckBox2 == null) {
			jCheckBox2 = new JCheckBox();
			jCheckBox2.setText("unzip *.jar from subdirs");
			jCheckBox2.setToolTipText("If checked, unpack Java Archive files from project-specific subdirs.");
		}
		return jCheckBox2;
	}

	/**
	 * This method initializes jButton16	
	 * 	
	 * @return javax.swing.JButton	
	 */
	private JButton getJButton16() {
		if (jButton16 == null) {
			jButton16 = new JButton();
			jButton16.setText("Open all");
			jButton16.setToolTipText("Open all projects simultaneously (ask before, *unpack if non-existing).");
			jButton16.addActionListener(new java.awt.event.ActionListener() {
				public void actionPerformed(java.awt.event.ActionEvent e) {
					if(jList.getModel().getSize()==0) return;
					int result = JOptionPane.showConfirmDialog(null, "Really open all projects now?", "Open all projects", JOptionPane.YES_NO_OPTION);
					if(result == JOptionPane.NO_OPTION || result == JOptionPane.CLOSED_OPTION) return;
					try {
						projectChangeMutex.acquire();
						projectChange.acquire();
						for(int i=0; i<jList.getModel().getSize(); i++) {
							try {
								File file = ((MyFile) jList.getModel().getElementAt(i)).getFile();
								myUnpack(file);
								
								// create BlueJ project file (if missing)
								FileHelpers.createBluejpkg(file);
								
								if(bluej!=null) {
									bluej.openProject(file);
								}
							} catch (Exception e1) {}
						}
					} catch (InterruptedException e1) {
					} finally {
						projectChange.release();
						projectChangeMutex.release();
					}
				}
			});
		}
		return jButton16;
	}

	/**
	 * This method initializes jButton17	
	 * 	
	 * @return javax.swing.JButton	
	 */
	private JButton getJButton17() {
		if (jButton17 == null) {
			jButton17 = new JButton();
			jButton17.setText("Clear");
			jButton17.setToolTipText("Delete ALL projects and files in workspace (ask first).");
			jButton17.addActionListener(new java.awt.event.ActionListener() {
				public void actionPerformed(java.awt.event.ActionEvent e) {
					if(JOptionPane.showConfirmDialog(null, "Really clear workspace?\nThis will delete ALL files and projects!", "Clear workspace", JOptionPane.OK_CANCEL_OPTION)
						== JOptionPane.OK_OPTION) {
						File workspace = new File(lastRefreshedWorkspace);
						boolean deleted = true;
						try {
							projectChangeMutex.acquire();
							projectChange.acquire();
							if(workspace.isDirectory()) {
								File[] files = workspace.listFiles();
								for(int i=0; i<files.length; i++) {
									if(files[i].isDirectory()) {
										boolean found = false;
										if(bluej!=null) found = myFindDelete(files[i]);
										if(!found) deleted &= FileHelpers.deleteDir(files[i]);
									}
									if(files[i].isFile()) {
										deleted &= files[i].delete();
									}
								}
							}
						} catch (InterruptedException e1) {
						} finally {
							projectChange.release();
							projectChangeMutex.release();
						}

						if(deleted) {
							myRefreshList();
						} else {
							myRefreshList();
							JOptionPane.showMessageDialog(null, "Could not clear workspace completely.");
						}
					}
					

				}
			});
		}
		return jButton17;
	}

	/**
	 * This method initializes jPanel19	
	 * 	
	 * @return javax.swing.JPanel	
	 */
	private JPanel getJPanel19() {
		if (jPanel19 == null) {
			jPanel19 = new JPanel();
			jPanel19.setLayout(new GridBagLayout());
		}
		return jPanel19;
	}

	/**
	 * This method initializes jCheckBox3	
	 * 	
	 * @return javax.swing.JCheckBox	
	 */
	private JCheckBox getJCheckBox3() {
		if (jCheckBox3 == null) {
			jCheckBox3 = new JCheckBox();
			jCheckBox3.setText("fixed position");
			jCheckBox3.setToolTipText("If checked, all newly opened package windows will keep the initial location.");
			jCheckBox3.addItemListener(new java.awt.event.ItemListener() {
				public void itemStateChanged(java.awt.event.ItemEvent e) {
					if(jCheckBox3.isSelected()) jCheckBox4.setSelected(false);
				}
			});
		}
		return jCheckBox3;
	}

	/**
	 * This method initializes jPanel20	
	 * 	
	 * @return javax.swing.JPanel	
	 */
	private JPanel getJPanel20() {
		if (jPanel20 == null) {
			jPanel20 = new JPanel();
			jPanel20.setLayout(new GridLayout(0,1));
			jPanel20.add(getJCheckBox5());
			jPanel20.add(getJCheckBox3());
			jPanel20.add(getJCheckBox4());
		}
		return jPanel20;
	}


	/**
	 * Calculate the position of a package window relative to the default package window.
	 * WARNING: This does not work from within BlueJ's PackageListener methods directly
	 * (as BlueJ sets the window position right after calling the packageOpened
	 * method).
	 * @param pkg The name of the package.
	 * @param location The location of the last package window.
	 * @param locationPkgName The name of the last package.
	 * @return The new location.
	 */
	private Point myCascadePackage(String pkgName, Point location, String locationPkgName) {
		Point newLocation = null;
		
		try {
			if(location==null || locationPkgName==null) return newLocation;
			
			int pkgLevel = 0;
			if(!locationPkgName.equals("")) pkgLevel++;
			for(int i=0; i<locationPkgName.length(); i++) {
				if(locationPkgName.charAt(i) == '.') pkgLevel++;
			}
			
			int level = 0;
			if(!pkgName.equals("")) level++;
			for(int i=0; i<pkgName.length(); i++) {
				if(pkgName.charAt(i) == '.') level++;
			}
			
			newLocation = new Point (
				Math.max(0, location.x - pkgLevel*20) + level*20,
				Math.max(0, location.y - pkgLevel*20) + level*20
			);
		} catch (Exception e) {
		}
		return newLocation;
	}

	/**
	 * This method initializes jCheckBox4	
	 * 	
	 * @return javax.swing.JCheckBox	
	 */
	private JCheckBox getJCheckBox4() {
		if (jCheckBox4 == null) {
			jCheckBox4 = new JCheckBox();
			jCheckBox4.setText("cascading position");
			jCheckBox4.setToolTipText("If checked, subpackage windows will be located relative to each other.");
			jCheckBox4.addItemListener(new java.awt.event.ItemListener() {
				public void itemStateChanged(java.awt.event.ItemEvent e) {
					if(jCheckBox4.isSelected()) jCheckBox3.setSelected(false);
				}
			});
		}
		return jCheckBox4;
	}

	/**
	 * This method initializes jCheckBox5	
	 * 	
	 * @return javax.swing.JCheckBox	
	 */
	private JCheckBox getJCheckBox5() {
		if (jCheckBox5 == null) {
			jCheckBox5 = new JCheckBox();
			jCheckBox5.setText("fixed size");
			jCheckBox5.setToolTipText("If checked, all newly opened package windows will keep the initial size.");
		}
		return jCheckBox5;
	}

	/**
	 * Main method for starting the GUI (without BlueJ).
	 * @param args ignored.
	 * @deprecated Use {@link MultiProject#main(String[])} instead.
	 */
	public static void main(String[] args) {
		(new GUI()).startupClosable();
	}
	
	/**
	 * Try to acquire a semaphore (will acquire only if not yet acquired,
	 * for JRE 1_4_2).
	 * @param sem The semaphore.
	 * @return true if acquired successfully, false otherwise.
	 */
	public static boolean tryAcquire(Semaphore sem) {
		boolean result = false;
		try {
			result = sem.attempt(0);
		} catch (InterruptedException e) {
			
		}
		return result;
	}
	
	
}  //  @jve:decl-index=0:visual-constraint="14,-6"
