package net.sf.multiproject;
import java.awt.event.ActionEvent;
import java.net.MalformedURLException;
//import java.util.concurrent.Semaphore;

import javax.swing.AbstractAction;
import javax.swing.JMenuItem;

import bluej.extensions.BPackage;
import bluej.extensions.BlueJ;
import bluej.extensions.Extension;
import bluej.extensions.MenuGenerator;
import bluej.extensions.event.ApplicationEvent;
import bluej.extensions.event.ApplicationListener;

/*
File: MultiProject.java
License: The MIT License

Copyright (c) 2008 Manuel Haim

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

/**
 * Multi-Project Workspace - Open and switch between multiple projects.
 * This is the main class.
 * @author Manuel Haim
 * @version 2008.06.25
 */
public class MultiProject extends Extension {

	/** Version string. */
	static final String MULTIPROJECT_VERSION = "2008.06.25";
	
	/** Name string. */
	static final String MULTIPROJECT_NAME = "Multi-Project Workspace";
	
	/** Tools menu item string. */
	static final String MULTIPROJECT_MENUITEM = "Multi-Project Workspace...";

	/** URL string. */
	private static final String MULTIPROJECT_URL = "http://multiproject.sf.net";
	
	/** Description string. */
	private static final String MULTIPROJECT_DESC = "A workspace for BlueJ which allows for easy importing/exporting and opening/switching between multiple projects.";

	/** The Multi-Project Workspace GUI. */
	private GUI gui;
	
	/** The BlueJ instance. */
	BlueJ bluej;

	/** A semaphore protecting the field {@link #started}. */
	static final Semaphore startedSem = new Semaphore(1);

	/** true if GUI is started, false otherwise. */
	static boolean started = false;

	/** A semaphore ensuring there is only one instance started. */
	static final Semaphore guiSem = new Semaphore(1);

	/** Property key for workspace path value. */
	static final String PROPERTY_WORKSPACE = "path";

	/** Property key for import path value. */
	static final String PROPERTY_IN = "importdir";

	/** Property key for boolean startup-with-bluej value. */
	static final String PROPERTY_STARTUP = "startup";
	
	/** Property key for import-files-from-subdirs checkbox value. */
	static final String PROPERTY_FILESFROMSUBDIRS = "filesfromsubdirs";
	
	/** Property key for import-jars-from-subdirs checkbox value. */
	static final String PROPERTY_JARSFROMSUBDIRS = "jarsfromsubdirs";

	/** Property key for exclude-javadoc checkbox value. */
	static final String PROPERTY_NOJAVADOC = "nojavadoc";
	
	/** Property key for fixed-window-size checkbox value. */
	static final String PROPERTY_FIXEDWINDOWSIZE = "fixedwindowsize";

	/** Property key for fixed-window-position checkbox value. */
	static final String PROPERTY_FIXEDWINDOWPOSITION = "fixedwindowpos";

	/** Property key for cascading-window-position checkbox value. */
	static final String PROPERTY_CASCADINGWINDOWPOSITION = "cascadingwindowpos";
	
	/** Property key for workspace-history-length value. */
	static final String PROPERTY_HISTORY_SIZE = "maxpaths";
	
	/** Filename for workspace settings. */
	static final String WORKSPACE_SETTINGS_FILE = "multiproject.ini";
	
	/** Filename for global settings (if started stand-alone without BlueJ). */
	static final String GLOBAL_SETTINGS_FILE = "multiproject_cfg.ini";
	
	/**
	 * The BlueJ tools menu item.
	 * @author Manuel Haim
	 */
	class MenuBuilder extends MenuGenerator {
	     public JMenuItem getToolsMenuItem(BPackage aPackage) {
	         return new JMenuItem(new SimpleAction(MULTIPROJECT_MENUITEM));
	     }
	
	     /*
	     public void notifyPostToolsMenu(BPackage bp, JMenuItem jmi) {
	     }
	     */
	     
	     class SimpleAction extends AbstractAction {
			private static final long serialVersionUID = 1L;
			public SimpleAction(String menuName) {
	             putValue(AbstractAction.NAME, menuName);
	         }
	         public void actionPerformed(ActionEvent anEvent) {
	        	 try {
	        		 guiSem.acquire();
	        		 if(gui==null) gui = new GUI();
	        		 gui.startup(bluej);
	        		 guiSem.release();
	        	 } catch (InterruptedException e) {}
	         }
	     }
	 }

	/**
	 * Main method for starting the Multi-Project Workspace without BlueJ.
	 * @param args
	 */
	public static void main(String[] args) {
		MultiProject mp = new MultiProject();
		try {
			guiSem.acquire();
			mp.gui = new GUI();
			mp.gui.startupClosable();
			guiSem.release();
		} catch (InterruptedException e) {}
		
	}

	public String getDescription() {
		return MULTIPROJECT_DESC;
	}

	public String getName() {
		return MULTIPROJECT_NAME;
	}

	public java.net.URL getURL() {
		java.net.URL result = null;
		try {
			result = new java.net.URL(MULTIPROJECT_URL);
		} catch (MalformedURLException e) {}
		return result;
	}

	public String getVersion() {
		return MULTIPROJECT_VERSION;
	}

	public boolean isCompatible() {
		return (VERSION_MAJOR>=2 && VERSION_MINOR>=4);
	}

	public void startup(BlueJ bluej) {
		this.bluej = bluej;
		MenuBuilder myMenus = new MenuBuilder();
		bluej.setMenuGenerator(myMenus);

		// Start GUI when BlueJ has finished startup.
		// (only if GUI was open on last BlueJ exit)
		bluej.addApplicationListener(new ApplicationListener() {
			public void blueJReady(ApplicationEvent arg0) {
				if(MultiProject.this.bluej.getExtensionPropertyString(PROPERTY_STARTUP, "").equals("true")) {
					try {
						guiSem.acquire();
						if(gui==null) gui = new GUI();
						gui.startup(MultiProject.this.bluej);
						guiSem.release();
					} catch (InterruptedException e) {}
					
				}
				MultiProject.this.bluej.removeApplicationListener(this);
			}
		});
		
	}

	public void terminate() {
		try {
			guiSem.acquire();
			if(gui!=null && !gui.isTerminated()) {
				gui.terminate();
				bluej.setExtensionPropertyString(PROPERTY_STARTUP, "true");
			}
			guiSem.release();
		} catch (InterruptedException e) {}
	}

}
